import os, sys, time
from itertools import combinations
import networkx as nx
import random
import math
import datetime


def index(i,j,k):
  return k*(i-1)+j
 
def solve_k_coloring(G, k, time_limit): 
   """
   Write CNF SAT file for k-coloring and call kissat
   """
   nvars    = G.number_of_nodes() * k  # for 1 \le i \le G.number_of_nodes() and 1 \le j \le k with index (k*(i-1) +j) refers vertex i being colored with color j
   nclauses = G.number_of_nodes() + G.number_of_edges()*k

   base_fname =  "color_" + str(k) +"_" +     datetime.datetime.now().strftime("%y%m%d_%H%M%S") 
   cnf_fname  = base_fname + ".cnf"
   res_fname  = base_fname + ".res"
   
   f = open(cnf_fname, "w")
   f.write("c testing coloring number " + str(k) + "\n")
   
   f.write("p cnf "+ str(nvars) + " "  + str(nclauses)+ "\n")

   for i in G.nodes():
       vertex_clause = ""
       for j in range(1,k+1):
           vertex_clause = vertex_clause + str(index(i,j,k)) + " " 

       vertex_clause = vertex_clause + "0\n"
       #f.write("c VCLS: " + vertex_clause)
       f.write(vertex_clause)

   
   for (u,v) in G.edges():
       assert(u <= G.number_of_nodes())
       assert(v <= G.number_of_nodes())
       for j in range(1,k+1):
           
           edge_clause = "-" + str(index(int(u),j,k)) + " -" + str(index(int(v),j,k)) + " 0\n"
           #f.write("c ECLS: " + edge_clause)
           f.write(edge_clause)
   
   f.close()

   cmd = "/opt/kissat/bin/kissat --time=" + time_limit + " "  + cnf_fname + " > " + res_fname
   print ("Executing `", cmd,"`")
   os.system(cmd)

   satisfiable = -1
   with open(res_fname,"r") as f:
       datafile = f.readlines()
   for line in datafile:
       if line.find("s UNSATISFIABLE") != -1:
           satisfiable = 0
           break 
       if line.find("s SATISFIABLE") != -1:
           satisfiable = 1
           break
            
   if satisfiable == -1:
       sys.exit("Error in kissat")

   return satisfiable

def color( G, PRINT=0, TLIM=600, solution=None): #{
    """
    SAT:
    ====
        Classic Assignment Formulation for Coloring


    PRINT is 3 Bit Number x2x1x0 \in { 0,..,7 }
        x0 :  Kissat Output	# 1
        x1 :  layout chanel     # 2
        x2 :  debug  chanel 1   # 4
        x3 :  debug  chanel 2   # 8

    """
    c00 = time.time()

    """ remove reflexive edges (v,v) \in E """
    for u,v in G.edges():
        if u == v:  G.remove_edge(u,v)

    V_0 = G.nodes()
    E_0 = G.edges()

    # We avoid the reduction 
    # G,dominator = coloring_preprocessing_fast( G )

    V = G.nodes()
    E = [ (u,v) for u,v in G.edges() if u != v ]

    if PRINT&4:
        print ('Reduction (V,E): (%s,%s) --> (%s,%s)    %0.2F sec  ' % (
                len(V_0), len(E_0), len(V), len(E), (time.time() - c00)));


    ub = len(set(nx.greedy_color(G).values()))
    #  _   _
    # | | | |
    # | |_| |
    # |  _  |
    # |_| |_|
    #
    H = ub

    if PRINT&4:
        print ('(V,E,H) = (%s,%s,%s)' % (len(V),len(E),ub))
    #   ___
    #  / _ \
    # | | | |
    # | |_| |
    #  \__\_\
    #
    c0 = time.time()
    Q = best_clique( G, ub )
    if PRINT&4:
        print ('Time Qq:', (time.time() - c0))

    lb = len(Q)

    if lb == ub:
        rt   = '%.2F' %(time.time() - c00)
        return len(V_0), len(E_0), len(V), len(E), len(Qq), H, lb, ub, 'init', rt

    if PRINT&2:
        print("")
        print ("Initial lb ", lb, "ub", ub)


    c0 = time.time()
    c_time_limit = c0+TLIM
    for j in list (reversed (range(lb,ub+1))):
        time_limit = c_time_limit - time.time()
        if (time_limit <= 0):
            print ("Time limit", TLIM,"reached")
            break
        if PRINT&2:
            print ("Testing coloring number ", j)
        k_colorable = solve_k_coloring(G, j, str(int(time_limit)))
        if k_colorable == 1:
            ub = j
        else:
            lb = j+1
            break 
            if PRINT&2:
                print ("Increased lb to ", lb)

    rtIP   = '%.2F' %(time.time() - c0)  
    rt   = '%.2F' %(time.time() - c00)  
    print("rt", rt)
    return len(V_0), len(E_0), len(V), len(E), lb, ub, lb, lb, rtIP, rt
#}

def best_clique( G, H  ): #{
    t0 = time.time()
    lenV = len(G.nodes())
    E = { (u,v) for u,v in G.edges() if u != v }
    Q=[]
    max_try=math.ceil(300*len(E)/lenV)
    Gcomplement = nx.complement( G )
    for i in range( max_try ):
        random.seed(i)
        _Q=nx.maximal_independent_set( Gcomplement )
        if len(Q) < len(_Q):
            Q = _Q
            string='\t|Q|: %s   %s sec\r' % ( len(Q), time.time() - t0 )
            sys.stdout.write( string );sys.stdout.flush()
        if time.time() - t0 > 60:
            break
    return Q
#}

def print_solution_to_check_sol_file(dominant_of,H,y,check_sol_file): #{
    if check_sol_file != None:
        epsilon=0.000000001
        sol=[-1]*(len(dominant_of)+1)
        for v in dominant_of:

            u = dominant_of[ v ]
            while u != dominant_of[ u ]:
                u = dominant_of[ u ]

            for i in range(0,H):
                if abs( y[u,i].x - 1 ) <= epsilon:
                    sol[ int(u) ] = i+1

            sol[ int(v) ] = sol[ int(u) ]

        os.system( 'echo -n > ' + check_sol_file )
        for c in sol:
            if c > -1:
                os.system( 'echo ' + str(c) + ' >> ' + check_sol_file )
#}

def coloring_preprocessing_fast( G ): #{
    """
        Removes dominated nodes.

        WARNING:
        ========
        This method are using the condition
                if adj[u] <= adj[v]:	# <=
        instead of
                if adj[u] < adj[v]:	# <
        and this is correct, iff graph has no reflexive edges (u,v) with u == v.
        Hence the reflexive edges must be deleted, before this method begins.

        EFFICIENCE:
        ==========
        This implementation is faster than coloring_preprocessing_old:

        coloring_preprocessing_old  Time(ash958GPIA.col) = 19.489439 sec
        coloring_preprocessing_fast Time(ash958GPIA.col) = 0.537492  sec
    """
    dominator = { v: v for v in G.nodes() }
    n=len(G.nodes())+1
    while len(G.nodes()) < n:
        n=len(G.nodes())
        adj={ v:set(G.neighbors(v)) for v in G }
        Vredundant = []
        for u,v in combinations( G.nodes(), 2):
            if adj[u] <= adj[v]:
                #print u, '<', v
                Vredundant.append( u )
                dominator[ u ]=v
            elif adj[v] <= adj[u]:
                #print v, '<', u
                Vredundant.append( v )
                dominator[ v ]=u
        G.remove_nodes_from(Vredundant)
    return G,dominator
#}
